#include <stdio.h>
#include "common.h"
#include "common_gl.h"

uint32_t window_width;
uint32_t window_height;

void common_gl_playback_init(
    VdpDevice           vdp_device,
    VdpGetProcAddress * vdp_get_proc_address,
    uint32_t            win_width,
    uint32_t            win_height
)
{
    int need_scratch_surface_buffer = 0;

    window_width = win_width;
    window_height = win_height;

    switch (opt_gold_surface_type) {
    case SURFACE_TYPE_GL_INTEROP_VID:
        need_scratch_surface_buffer = 1;
        break;
    case SURFACE_TYPE_GL_WIN:
        surface_buffer_size = window_width * window_height * 4;
        break;
    default:
        break;
    }

    common_playback_init(vdp_device, vdp_get_proc_address, need_scratch_surface_buffer);
}

void common_gl_do_gold_gl_tex_yuv_nv12x2_as_nv12(uint32_t frame_number, GLuint target, GLuint * texs)
{
    int i;

    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    uint8_t *buffer = scratch_surface_buffer;
    size_t texsize = trace_stream_conf_ptr->vid_width * trace_stream_conf_ptr->vid_height / 2;
    GLenum format = GL_RED;

    for (i = 0; i < 4; ++i) {
        glBindTexture(target, texs[i]);
        glGetTexImage(target, 0, format, GL_UNSIGNED_BYTE, buffer);
        buffer += texsize;
        if (i == 1) {
            texsize /= 2;
            format = GL_RG;
        }
    }
    glBindTexture(target, 0);

    common_nv12x2_to_nv12(trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, scratch_surface_buffer, surface_buffer);

    common_do_gold_buffer(frame_number, "nv12", trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, surface_buffer, surface_buffer_size);
}

void common_gl_do_gold_gl_tex_rgb(uint32_t frame_number, GLuint target, GLuint tex)
{
    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    glBindTexture(target, tex);
    glGetTexImage(target, 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, surface_buffer);
    glBindTexture(target, 0);

    common_do_gold_buffer_argb_to_rgb(frame_number, trace_stream_conf_ptr->out_width, trace_stream_conf_ptr->out_height, surface_buffer, surface_buffer_size, 0);
}

void common_gl_do_gold_window(uint32_t frame_number)
{
    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    glReadPixels(0, 0, window_width, window_height, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, surface_buffer);

    common_do_gold_buffer_argb_to_rgb(frame_number, window_width, window_height, surface_buffer, surface_buffer_size, 1);
}

void common_gl_playback_fini()
{
    common_playback_fini();
}

