#!/usr/bin/env python

import os
import os.path
import re
import sys

rec_tagline = re.compile("^traceplay gold (\\w+) (\\w+) (\\d+) (\\d+) content_size=(\\d+)$")

app = sys.argv.pop(0)
in_gold_fn = sys.argv.pop(0)
out_gold_fn = sys.argv.pop(0)

# The following is basically just:
#     grep -v 'traceplay gold' < $in_gold_fn > $out_gold_fn
# However, the gold data (for 'content' at least) doesn't end in \n,
# and since grep is line-oriented, the above simple solution doesn't work.

try:
    in_f = file(in_gold_fn, "rb")
except:
    print >>sys.stderr, "Can't open input file '%s'" % in_gold_fn
    print >>sys.stderr, ""
    raise

try:
    out_f = file(out_gold_fn, "wb")
except:
    print >>sys.stderr, "Can't open output file '%s'" % out_gold_fn
    print >>sys.stderr, ""
    raise

while True:
    tagline = in_f.readline()
    tagline.strip()
    if not tagline:
        break

    match = rec_tagline.match(tagline)
    if not match:
        print >>sys.stderr, "tagline does not match format:"
        print >>sys.stderr, "    " + tagline
        sys.exit(1)

    if match.group(1) != 'content':
        print >>sys.stderr, "File contains '%s' not 'content'" % match.group(1)
        sys.exit(1)

    if match.group(2) != 'nv12':
        print >>sys.stderr, "File is '%s' not 'nv12'" % match.group(2)
        sys.exit(1)

    width = match.group(3)
    height = match.group(4)

    try:
        content_size = int(match.group(5))
    except:
        print >>sys.stderr, "Can't convert content_size '%s' to integer" % match.group(1)
        print >>sys.stderr, ""
        raise

    content = in_f.read(content_size)

    out_f.write(content)

out_f.close()
in_f.close()

print "To play back the result, use:"
print "mplayer -demuxer rawvideo -rawvideo format=nv12:w=%s:h=%s %s" % (width, height, out_gold_fn)

