/*
 * Copyright (c) 2008-2009 NVIDIA Corporation
 *
 * Based on sample OpenGL code bearing this notice:
 *
 * This code was created by Jeff Molofee '99 
 * (ported to Linux/GLX by Mihael Vrbanec '00)
 *
 * If you've found this code useful, please let me know.
 *
 * Visit Jeff at http://nehe.gamedev.net/
 * 
 * or for port-specific comments, questions, bugreports etc. 
 * email to Mihael.Vrbanec@stud.uni-karlsruhe.de
 */
 
#include <stdio.h>
#include <math.h>
#include <vdpau/vdpau_x11.h>
#include "traceplay.h"
#include "common.h"
#include "common_gl.h"

VdpDevice           vdp_device;
VdpGetProcAddress * vdp_get_proc_address;

VdpDeviceDestroy *                       vdp_device_destroy;
VdpPresentationQueueTargetDestroy *      vdp_presentation_queue_target_destroy;
VdpPresentationQueueCreate *             vdp_presentation_queue_create;
VdpPresentationQueueDestroy *            vdp_presentation_queue_destroy;
VdpPresentationQueueDisplay *            vdp_presentation_queue_display;
VdpPresentationQueueQuerySurfaceStatus * vdp_presentation_queue_query_surface_status;
VdpPresentationQueueTargetCreateX11 *    vdp_presentation_queue_target_create_x11;

VdpPresentationQueueTarget vdp_flip_target;
VdpPresentationQueue       vdp_flip_queue;

/* stuff about our window grouped together */
typedef struct {
    Display *dpy;
    int screen;
    Window win;
    GLXContext ctx;
    XSetWindowAttributes attr;
    int x, y;
    unsigned int width, height;
    unsigned int depth;    
    Pixmap pixmap;
    GLXPixmap glxpixmap;
    GLuint tex;
} GLWindow;

static int attrList[] = {
    GLX_DOUBLEBUFFER, True, // createGLWindow assumes this is first
    GLX_X_RENDERABLE, True,
    GLX_X_VISUAL_TYPE, GLX_TRUE_COLOR,
    GLX_RENDER_TYPE, GLX_RGBA_BIT,
    GLX_RED_SIZE, 8,
    GLX_GREEN_SIZE, 8,
    GLX_BLUE_SIZE, 8,
    GLX_ALPHA_SIZE, 8,
    GLX_DEPTH_SIZE, 16,
    GLX_DRAWABLE_TYPE, (GLX_PIXMAP_BIT | GLX_WINDOW_BIT),
    GLX_BIND_TO_TEXTURE_TARGETS_EXT, True,
    GLX_BIND_TO_TEXTURE_RGBA_EXT, True,
    GLX_Y_INVERTED_EXT, True,
    None
};

GLWindow GLWin;
GLfloat rotX = 0.6;  /* X Rotation */
GLfloat rotY = 0.8;  /* Y Rotation */
GLfloat rotZ = 1.2;  /* Z Rotation */

/* function called when our window is resized (should only happen in window mode) */
void resizeGLScene(unsigned int width, unsigned int height)
{
    if (height == 0)    /* Prevent A Divide By Zero If The Window Is Too Small */
        height = 1;
    /* Reset The Current Viewport And Perspective Transformation */
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (GLfloat)width / (GLfloat)height, 0.1f, 100.0f);
    glMatrixMode(GL_MODELVIEW);
}

/* general OpenGL initialization function */
void initGL(GLvoid)
{
    glGenTextures(1, &GLWin.tex);
    glBindTexture(GL_TEXTURE_2D, GLWin.tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glEnable(GL_TEXTURE_2D);
    glShadeModel(GL_SMOOTH);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    /* we use resizeGLScene once to set up our initial perspective */
    resizeGLScene(GLWin.width, GLWin.height);
    glFlush();
}

void * traceplay_video_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface
)
{
    return 0;
}

void * traceplay_output_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface
)
{
    return 0;
}

void traceplay_video_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
}

void traceplay_output_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
}

void traceplay_display_video_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
    printf("ERROR at %s:%d: display_video_surface called\n", __FILE__, __LINE__);
    common_exit_error();
}

void traceplay_wait_for_output_surface_idle(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
}

void drawGLScene(uint32_t frame_number, GLuint tex);

void traceplay_display_output_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    vdp_presentation_queue_display(vdp_flip_queue, surface, 0, 0, 0);
    for (;;) {
        VdpPresentationQueueStatus status;
        VdpTime                    first_presentation_time;

        vdp_presentation_queue_query_surface_status(
            vdp_flip_queue,
            surface,
            &status,
            &first_presentation_time
        );
        if (status == VDP_PRESENTATION_QUEUE_STATUS_VISIBLE) {
            break;
        }
    }

    glBindTexture(GL_TEXTURE_2D, GLWin.tex);
    glXBindTexImageEXT(GLWin.dpy, GLWin.glxpixmap, GLX_FRONT_LEFT_EXT, NULL);
    glBindTexture(GL_TEXTURE_2D, 0);

    if (opt_gold_surface_type == SURFACE_TYPE_GL_TFP_OUT) {
        common_gl_do_gold_gl_tex_rgb(frame_number, GL_TEXTURE_2D, GLWin.tex);
    }

    drawGLScene(frame_number, GLWin.tex);

    glBindTexture(GL_TEXTURE_2D, GLWin.tex);
    glXReleaseTexImageEXT(GLWin.dpy, GLWin.glxpixmap, GLX_FRONT_LEFT_EXT);
    glBindTexture(GL_TEXTURE_2D, 0);
}

void traceplay_process_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        vid_surface,
    void *                 vid_surface_context,
    VdpOutputSurface       out_surface,
    void *                 out_surface_context
)
{
    printf("ERROR at %s:%d: process_surface called\n", __FILE__, __LINE__);
    common_exit_error();
}

Traceplay_Callbacks tpcb = {
    traceplay_video_surface_post_create,
    traceplay_output_surface_post_create,
    traceplay_video_surface_pre_destroy,
    traceplay_output_surface_pre_destroy,
    traceplay_display_video_surface,
    traceplay_wait_for_output_surface_idle,
    traceplay_display_output_surface,
    traceplay_process_surface
};

void initVDPAU(void)
{
    VdpStatus vdp_st;

    // Create Device

    vdp_st = vdp_device_create_x11(
        GLWin.dpy,
        GLWin.screen,
        &vdp_device,
        &vdp_get_proc_address
    );
    CHECK_ST

    // Entry point retrieval

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_DEVICE_DESTROY,
        (void **)&vdp_device_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_TARGET_DESTROY,
        (void **)&vdp_presentation_queue_target_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_CREATE,
        (void **)&vdp_presentation_queue_create
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_DESTROY,
        (void **)&vdp_presentation_queue_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_DISPLAY,
        (void **)&vdp_presentation_queue_display
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_QUERY_SURFACE_STATUS,
        (void **)&vdp_presentation_queue_query_surface_status
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_TARGET_CREATE_X11,
        (void **)&vdp_presentation_queue_target_create_x11
    );
    CHECK_ST

    vdp_st = vdp_presentation_queue_target_create_x11(
        vdp_device,
        GLWin.pixmap,
        &vdp_flip_target
    );
    CHECK_ST

    vdp_st = vdp_presentation_queue_create(
        vdp_device,
        vdp_flip_target,
        &vdp_flip_queue
    );
    CHECK_ST
}

void finiVDPAU(void)
{
    vdp_presentation_queue_destroy(
        vdp_flip_queue
    );

    vdp_presentation_queue_target_destroy(
        vdp_flip_target
    );

    vdp_device_destroy(
        vdp_device
    );
}

/* Here goes our drawing code */
void drawGLScene(uint32_t frame_number, GLuint tex)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    glTranslatef(0.0f, 0.0f, -5.0f);
    glRotatef(rotX, 1.0f, 0.0f, 0.0f);  /* rotate on the X axis */
    glRotatef(rotY, 0.0f, 1.0f, 0.0f);  /* rotate on the Y axis */
    glRotatef(rotZ, 0.0f, 0.0f, 1.0f);  /* rotate on the Z axis */

    glBindTexture(GL_TEXTURE_2D, tex);

#define glTexCoord2f(x, y) glTexCoord2f((x), (1.0 - (y)));

    glBegin(GL_QUADS);
        /* front face */
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, 1.0f); 
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, 1.0f);
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, 1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, 1.0f);
        /* back face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, -1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, -1.0f);
        /* right face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, -1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, 1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, 1.0f);
        /* left face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, 1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, 1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, -1.0f);
        /* top face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, 1.0f, 1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0f, 1.0f, -1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0f, 1.0f, 1.0f);
        /* bottom face */
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, -1.0f); 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0f, -1.0f, 1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0f, -1.0f, 1.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0f, -1.0f, -1.0f);
    glEnd();

    if (opt_gold_surface_type == SURFACE_TYPE_GL_WIN) {
        common_gl_do_gold_window(frame_number);
    }

    glXSwapBuffers(GLWin.dpy, GLWin.win);

    /* change the rotation angles */
    rotX += 0.3f;
    rotY += 0.2f;
    rotZ += 0.4f;
}

/* function to release/destroy our resources and restoring the old desktop */
GLvoid killGLWindow(GLvoid)
{
    if (GLWin.ctx) {
        if (!glXMakeCurrent(GLWin.dpy, None, NULL)) {
            printf("Could not release drawing context.\n");
        }
        glXDestroyContext(GLWin.dpy, GLWin.ctx);
        GLWin.ctx = NULL;
    }
    XCloseDisplay(GLWin.dpy);
}

/* this function creates our window and sets it up properly */
/* FIXME: bits is currently unused */
void createGLWindow(char* title, int width, int height, int bits)
{
    XVisualInfo *vi;
    Colormap cmap;
    int glxMajorVersion, glxMinorVersion;
    Atom wmDelete;
    Window winDummy;
    unsigned int borderDummy;
    int nfbconfigs, i;
    GLXFBConfig *fbconfigs;
 
    /* get a connection */
    GLWin.dpy = XOpenDisplay(0);
    if (!GLWin.dpy) {
        printf("ERROR at %s:%d: XOpenDisplay failed\n", __FILE__, __LINE__);
        common_exit_error();
    }
    GLWin.screen = DefaultScreen(GLWin.dpy);

    fbconfigs = glXChooseFBConfig(GLWin.dpy, GLWin.screen, attrList, &nfbconfigs);
    if (!fbconfigs) {
        attrList[1] = False;
        fbconfigs = glXChooseFBConfig(GLWin.dpy, GLWin.screen, attrList, &nfbconfigs);
        if (!fbconfigs) {
            printf("ERROR at %s:%d: glXChooseFBConfig returned nothing\n", __FILE__, __LINE__);
            common_exit_error();
        }
    }
    i = 0;
    vi = glXGetVisualFromFBConfig(GLWin.dpy, fbconfigs[i]);

    glXQueryVersion(GLWin.dpy, &glxMajorVersion, &glxMinorVersion);
    printf("glX-Version %d.%d\n", glxMajorVersion, glxMinorVersion);

    /* create a GLX context */
    GLWin.ctx = glXCreateContext(GLWin.dpy, vi, 0, GL_TRUE);

    /* create a color map */
    cmap = XCreateColormap(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
        vi->visual, AllocNone);
    GLWin.attr.colormap = cmap;
    GLWin.attr.border_pixel = 0;

    /* create a window in window mode*/
    GLWin.attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask |
        StructureNotifyMask;
    GLWin.win = XCreateWindow(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
        0, 0, width, height, 0, vi->depth, InputOutput, vi->visual,
        CWBorderPixel | CWColormap | CWEventMask, &GLWin.attr);
    GLWin.pixmap = XCreatePixmap(GLWin.dpy, GLWin.win,
        trace_stream_conf_ptr->out_width, trace_stream_conf_ptr->out_height, vi->depth);

    /* only set window title and handle wm_delete_events if in windowed mode */
    wmDelete = XInternAtom(GLWin.dpy, "WM_DELETE_WINDOW", True);
    XSetWMProtocols(GLWin.dpy, GLWin.win, &wmDelete, 1);
    XSetStandardProperties(GLWin.dpy, GLWin.win, title,
        title, None, NULL, 0, NULL);
    XMapRaised(GLWin.dpy, GLWin.win);

    /* connect the glx-context to the window */
    glXMakeCurrent(GLWin.dpy, GLWin.win, GLWin.ctx);

    int pixmapAttribs[] = {
        GLX_TEXTURE_TARGET_EXT, GLX_TEXTURE_2D_EXT,
        GLX_TEXTURE_FORMAT_EXT, GLX_TEXTURE_FORMAT_RGBA_EXT,
        None
    };
    GLWin.glxpixmap = glXCreatePixmap(GLWin.dpy, fbconfigs[i], GLWin.pixmap, pixmapAttribs);

    XGetGeometry(GLWin.dpy, GLWin.win, &winDummy, &GLWin.x, &GLWin.y,
        &GLWin.width, &GLWin.height, &borderDummy, &GLWin.depth);
    printf("Depth %d\n", GLWin.depth);

    if (glXIsDirect(GLWin.dpy, GLWin.ctx)) 
        printf("Congrats, you have Direct Rendering!\n");
    else
        printf("Sorry, no Direct Rendering possible!\n");

    initGL(); 
}

void help()
{
    printf("app [options]\n");
    common_print_help();
}

int main(int argc, char * * argv)
{
    int ret;

    app_name = "traceplay_gl_tfp";
    support_opt_gold_surface_type_mask =
        (1 << SURFACE_TYPE_VDP_VID_DEC_ORDER) |
        (1 << SURFACE_TYPE_VDP_VID_DISP_ORDER) |
        (1 << SURFACE_TYPE_VDP_OUT) |
        (1 << SURFACE_TYPE_GL_TFP_OUT) |
        (1 << SURFACE_TYPE_GL_WIN);

    common_init();

    argc--;
    argv++;
    while (argc) {
        if ((ret = common_parse_arg(&argc, &argv)) != COMMON_PARSE_ARG_NOT_PARSED) {
            if (ret != COMMON_PARSE_ARG_OK) {
                help();
                common_exit_error();
            }
        }
        else {
            printf("ERROR: Unknown option '%s'\n", argv[0]);
            help();
            common_exit_error();
        }
        argc--;
        argv++;
    }

    if (common_check_args() != COMMON_CHECK_ARGS_OK) {
        help();
        common_exit_error();
    }

    common_load_stream();

#define WIN_WIDTH  1280
#define WIN_HEIGHT 1024
    createGLWindow(
        "VDPAU/GL Interop (GLX_EXT_texture_from_pixmap)",
        WIN_WIDTH,
        WIN_HEIGHT,
        24
    );
    initVDPAU();
    common_gl_playback_init(
        vdp_device,
        vdp_get_proc_address,
        WIN_WIDTH,
        WIN_HEIGHT
    );
    struct Traceplay_Ctx *traceplay_ctx = traceplay_init(
        vdp_device,
        vdp_get_proc_address,
        TRACEPLAY_PLAY_TYPE_OUTPUT,
        &tpcb,
        trace_stream_conf_ptr
    );
    CHECK_PTR(traceplay_ctx)
    traceplay_play(
        traceplay_ctx,
        trace_cmds_ptr,
        opt_repeat_count
    );
    traceplay_fini(
        traceplay_ctx
    );
    common_gl_playback_fini();
    finiVDPAU();
    killGLWindow();

    common_exit_ok();
    return 0;
}

