#!/usr/bin/env python

import cStringIO
import re
import sys

app = sys.argv.pop(0)
fntrace = sys.argv.pop(0)

dec_id = 0
disp_id = 0
vid_surface_size = None
out_surface_size = None
vid_surface_new_id = 0
vid_surfaces = {}
decoder_profile = None
decoder_max_references = None
cmd_deps = cStringIO.StringIO()
cmds = cStringIO.StringIO()

def split_braces(l):
    fields = []

    pos = 0
    maxpos = len(l)
    while pos < maxpos:
        lbracepos = l.find('{', pos)
        if lbracepos < 0:
            lbracepos = maxpos + 1
        commapos = l.find(',', pos)
        if commapos < 0:
            commapos = maxpos
        #print "top pos:", lbracepos, commapos
        if commapos < lbracepos:
            fields.append(l[pos:commapos].strip())
            pos = commapos + 1
            #print 'comma first, pos now', pos, 'got field', repr(fields[-1])
        else:
            pos2 = lbracepos + 1
            bracedepth = 1
            lbracepos = l.find('{', pos2)
            rbracepos = l.find('}', pos2)
            while (pos2 < maxpos) and bracedepth:
                #print 'matching', lbracepos, rbracepos
                if (lbracepos >= 0) and (lbracepos < rbracepos):
                    bracedepth += 1
                    pos2 = lbracepos + 1
                    lbracepos = l.find('{', pos2)
                elif (rbracepos >= 0):
                    bracedepth -= 1
                    pos2 = rbracepos + 1
                    rbracepos = l.find('}', pos2)
                else:
                    pos2 = maxpos
            fields.append(l[pos:pos2].strip())
            pos = pos2
            commapos = l.find(',', pos)
            if commapos > 0:
                pos = commapos + 1
            #print 'brace first, pos now', pos, 'got field', repr(fields[-1])

    return fields

def vdp_video_surface_create(l):
    global vid_surface_size

    fields = l.split(',')
    w = fields[2].strip()
    h = fields[3].strip()

    size = (w, h)
    if not vid_surface_size:
        vid_surface_size = size
    else:
        if size != vid_surface_size:
            raise Exception('Inconsistent video surface size')

def vdp_output_surface_create(l):
    global out_surface_size

    fields = l.split(',')
    w = fields[2].strip()
    h = fields[3].strip()
    size = (w, h)
    if not out_surface_size:
        out_surface_size = size
    else:
        if size != out_surface_size:
            raise Exception('Inconsistent output surface size')

pi_type_map = {
  0:  ("VdpPictureInfoMPEG1Or2", False),
  1:  ("VdpPictureInfoMPEG1Or2", False),
  2:  ("VdpPictureInfoMPEG1Or2", False),
  6:  ("VdpPictureInfoH264", True),
  7:  ("VdpPictureInfoH264", True),
  8:  ("VdpPictureInfoH264", True),
  9:  ("VdpPictureInfoVC1", False),
  10: ("VdpPictureInfoVC1", False),
  11: ("VdpPictureInfoVC1", False),
  12: ("VdpPictureInfoMPEG4Part2", False),
  13: ("VdpPictureInfoMPEG4Part2", False),
  14: ("VdpPictureInfoMPEG4Part2", False),
  15: ("VdpPictureInfoMPEG4Part2", False),
  16: ("VdpPictureInfoMPEG4Part2", False),
  17: ("VdpPictureInfoMPEG4Part2", False),
  18: ("VdpPictureInfoMPEG4Part2", False),
  19: ("VdpPictureInfoMPEG4Part2", False),
  20: ("VdpPictureInfoMPEG4Part2", False),
  21: ("VdpPictureInfoMPEG4Part2", False),
}

def vdp_decoder_create(l):
    global decoder_profile
    global decoder_max_references

    fields = l.split(',')
    decoder_profile = int(fields[1].strip())
    decoder_max_references = fields[4].strip()

    if not pi_type_map.has_key(decoder_profile):
        raise Exception('Unsupported decoder profile')

bb_ver_prefix = '(ver 0) '
len_bb_ver_prefix = len(bb_ver_prefix)

def map_vid_surf(s):
    if s == '4294967295':
        s = 'VDP_INVALID_HANDLE'
    else:
        if not vid_surfaces.has_key(s):
            raise Exception('Reference is undecoded surface')
        s = str(vid_surfaces[s])
    return s

def vdp_decoder_render(l):
    global vid_surface_new_id
    global dec_id

    this_dec_id = dec_id
    dec_id += 1

    fields = split_braces(l)
    target = fields[1]
    picture_info = fields[2]
    bitstream_buffer_count = int(fields[3])
    bitstream_buffers_txt = fields[4]

    if not vid_surfaces.has_key(target):
        vid_surfaces[target] = vid_surface_new_id
        vid_surface_new_id += 1
    vid_surface_index = vid_surfaces[target]

    (pi_type, is_h264) = pi_type_map.get(decoder_profile)

    pi_fields = split_braces(picture_info[1:-1])
    if is_h264:
        ref_frames = split_braces(pi_fields[29][1:-1])
        ref_frames_new = []
        for i in range(16):
            rf_fields = split_braces(ref_frames[i][1:-1])
            rf_fields[0] = map_vid_surf(rf_fields[0])
            ref_frames_new.append('{' + ', '.join(rf_fields) + '}')
        pi_fields[29] = '{' + ', '.join(ref_frames_new) + '}'
    else:
        for i in range(2):
            pi_fields[i] = map_vid_surf(pi_fields[i])
    picture_info = '{' + ', '.join(pi_fields) + '}'

    print >>cmd_deps, '''
%(pi_type)s picinfo_%(this_dec_id)d = %(picture_info)s;
''' % locals()

    bb_lens = []
    bitstream_buffers = split_braces(bitstream_buffers_txt[1:-1])
    if len(bitstream_buffers) != bitstream_buffer_count:
        raise Exception('Inconsistent # of bitstream buffers')
    for i in range(bitstream_buffer_count):
        bb_fields = split_braces(bitstream_buffers[i][1:-1])
        if not bb_fields[0].startswith(bb_ver_prefix):
            raise Exception('Unknown bitstream buffer version')
        bb_lens.append(bb_fields[0][len_bb_ver_prefix:])
        bb_data = (' ' + bb_fields[1]).replace(' ', ', 0x')[2:]

        print >>cmd_deps, '''\
uint8_t bitstream_%(this_dec_id)d_%(i)d[] = {''' % locals()

        pos = 0
        len_bb_data = len(bb_data)
        while pos < len_bb_data:
            t = bb_data[pos:pos+96]
            print >>cmd_deps, '   ', t
            pos += 96

        print >>cmd_deps, '''\
};
''' % locals()

    print >>cmd_deps, '''\
VdpBitstreamBuffer bitstream_buffers_%(this_dec_id)d[] = {''' % locals()

    for i in range(bitstream_buffer_count):
        bb_len = bb_lens[i]
        print >>cmd_deps, '''\
    {
        0,
        bitstream_%(this_dec_id)d_%(i)d,
        %(bb_len)s
    }, ''' % locals()

    print >>cmd_deps, '''\
};

Traceplay_CmdDecode cmd_dec_%(this_dec_id)d = {
    %(vid_surface_index)d,
    (VdpPictureInfo *)&picinfo_%(this_dec_id)d,
    %(bitstream_buffer_count)d,
    bitstream_buffers_%(this_dec_id)d
};''' % locals()

    print >>cmds, '''\
    {TRACEPLAY_CMD_DECODE,  &cmd_dec_%(this_dec_id)d},''' % locals()

def vdp_video_mixer_render(l):
    global disp_id

    this_disp_id = disp_id
    disp_id += 1

    fields = split_braces(l)
    target = fields[6]

    if not vid_surfaces.has_key(target):
        raise Exception('Display undecoded surface')
    vid_surface_index = vid_surfaces[target]

    print >>cmd_deps, '''\
Traceplay_CmdDisplay cmd_disp_%(this_disp_id)d = {
    %(vid_surface_index)d
};''' % locals()

    print >>cmds, '''\
    {TRACEPLAY_CMD_DISPLAY, &cmd_disp_%(this_disp_id)s},''' % locals()

funcs = {
    'vdp_video_surface_create':  vdp_video_surface_create,
    'vdp_output_surface_create': vdp_output_surface_create,
    'vdp_decoder_create':        vdp_decoder_create,
    'vdp_decoder_render':        vdp_decoder_render,
    'vdp_video_mixer_render':    vdp_video_mixer_render
}

f = file(fntrace, "rt")
for l in f:
    bracket_pos = l.find('(')
    if bracket_pos <= 0:
        continue
    cmd = l[:bracket_pos]
    func = funcs.get(cmd, None)
    if not func:
        continue
    func(l[bracket_pos+1:-1])
f.close()

vid_surface_width = vid_surface_size[0]
vid_surface_height = vid_surface_size[1]
vid_surface_count = len(vid_surfaces.keys())
out_surface_width = out_surface_size[0]
out_surface_height = out_surface_size[1]
cmd_deps_txt = cmd_deps.getvalue()
cmds_txt = cmds.getvalue()

print '''\
#include "traceplay.h"

Traceplay_Stream_Conf traceplay_stream_conf = {
    %(vid_surface_width)s,
    %(vid_surface_height)s,
    %(vid_surface_count)d,
    %(out_surface_width)s,
    %(out_surface_height)s,
    (VdpDecoderProfile)%(decoder_profile)s,
    %(decoder_max_references)s
};
%(cmd_deps_txt)s
Traceplay_Cmds traceplay_cmds_array[] = {
%(cmds_txt)s
    {TRACEPLAY_CMD_END,     NULL}
};

Traceplay_Cmds * traceplay_cmds = &traceplay_cmds_array[0];
''' % locals()

